/*
 * Copyright (C) 2023 Xeme Developers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <https://www.gnu.org/licenses/>.
 */

#include "xememessage.h"
#include "xemestring.h"

enum {
	PROP_0,
	PROP_THREAD,
	PROP_THREAD_PARENT,
	N_PROPERTIES,
};
static GParamSpec *properties[N_PROPERTIES] = {NULL, };

struct _XemeMessage {
	XemeStanza parent;

	GHashTable *bodies;
	GPtrArray *bodies_order;

	GHashTable *subjects;
	GPtrArray *subjects_order;

	char *thread;
	char *thread_parent;
};

/******************************************************************************
 * Helpers
 *****************************************************************************/
static void
xeme_message_add_element_with_language(XemeMessage *message,
                                       GHashTable *collection,
                                       GPtrArray *order,
                                       const char *language,
                                       const char *content)
{
	char *lang = NULL;
	gboolean found = FALSE;
	gpointer key = NULL;

	g_return_if_fail(XEME_IS_MESSAGE(message));
	g_return_if_fail(content != NULL);

	/* If the caller passes in NULL, we just convert it to empty string. */
	if(!xeme_str_is_empty(language)) {
		lang = g_strdup(language);
	} else {
		lang = g_strdup("");
	}

	/* Look if the language already exists, and if so, remove it from our
	 * ordered list of languages.
	 */
	found = g_hash_table_lookup_extended(collection, lang, &key, NULL);
	if(found) {
		g_ptr_array_remove(order, key);
	}

	/* Add the language to our list of ordered languages. This is used to look
	 * up keys in the bodies hash table, so we just keep a pointer to that
	 * value rather than creating a new copy and requiring a strcmp to find it
	 * again.
	 */
	g_ptr_array_add(order, lang);

	/* We use g_hash_table_replace because we want to free the old key and
	 * have it use the new one we pass in, which is the opposite of what
	 * g_hash_table_insert does.
	 */
	g_hash_table_replace(collection, lang, g_strdup(content));
}

static void
xeme_message_marshal_language_collection(GHashTable *collection,
                                         GPtrArray *order,
                                         GString *str,
                                         const char *element)
{
	for(guint i = 0; i < order->len; i++) {
		const char *lang = NULL;
		const char *content = NULL;

		lang = g_ptr_array_index(order, i);
		content = g_hash_table_lookup(collection, lang);

		g_string_append_printf(str, "<%s", element);
		if(xeme_str_is_empty(lang)) {
			g_string_append_printf(str, ">%s", content);
		} else {
			g_string_append_printf(str, " xml:lang=\"%s\">%s", lang, content);
		}
		g_string_append_printf(str, "</%s>", element);
	}
}

/******************************************************************************
 * XemeStanza Implementation
 *****************************************************************************/
static char *
xeme_message_marshal(XemeStanza *stanza, G_GNUC_UNUSED GError **error) {
	XemeMessage *message = XEME_MESSAGE(stanza);
	GString *str = NULL;
	char *attrs = NULL;

	str = g_string_new("<message");

	attrs = xeme_stanza_marshal_attributes(stanza);
	if(!xeme_str_is_empty(attrs)) {
		g_string_append_printf(str, " %s>", attrs);
	} else {
		g_string_append(str, ">");
	}
	g_clear_pointer(&attrs, g_free);

	/* Now run through the collections. */
	xeme_message_marshal_language_collection(message->subjects,
	                                         message->subjects_order, str,
	                                         "subject");
	xeme_message_marshal_language_collection(message->bodies,
	                                         message->bodies_order, str,
	                                         "body");

	if(!xeme_str_is_empty(message->thread)) {
		if(!xeme_str_is_empty(message->thread_parent)) {
			g_string_append_printf(str, "<thread parent=\"%s\">%s</thread>",
			                       message->thread_parent, message->thread);
		} else {
			g_string_append_printf(str, "<thread>%s</thread>",
			                       message->thread);
		}
	}

	/* Finish up */
	g_string_append(str, "</message>");

	return g_string_free_and_steal(str);
}

/******************************************************************************
 * GObject Implementation
 *****************************************************************************/
G_DEFINE_FINAL_TYPE(XemeMessage, xeme_message, XEME_TYPE_STANZA)

static void
xeme_message_finalize(GObject *obj) {
	XemeMessage *message = XEME_MESSAGE(obj);

	g_ptr_array_free(message->subjects_order, TRUE);
	message->subjects_order = NULL;
	g_clear_pointer(&message->subjects, g_hash_table_destroy);

	g_ptr_array_free(message->bodies_order, TRUE);
	message->bodies_order = NULL;
	g_clear_pointer(&message->bodies, g_hash_table_destroy);

	g_clear_pointer(&message->thread, g_free);
	g_clear_pointer(&message->thread_parent, g_free);

	G_OBJECT_CLASS(xeme_message_parent_class)->finalize(obj);
}

static void
xeme_message_get_property(GObject *obj, guint param_id, GValue *value,
                          GParamSpec *pspec)
{
	XemeMessage *message = XEME_MESSAGE(obj);

	switch(param_id) {
	case PROP_THREAD:
		g_value_set_string(value, xeme_message_get_thread(message));
		break;
	case PROP_THREAD_PARENT:
		g_value_set_string(value, xeme_message_get_thread_parent(message));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_message_set_property(GObject *obj, guint param_id, const GValue *value,
                          GParamSpec *pspec)
{
	XemeMessage *message = XEME_MESSAGE(obj);

	switch(param_id) {
	case PROP_THREAD:
		xeme_message_set_thread(message, g_value_get_string(value));
		break;
	case PROP_THREAD_PARENT:
		xeme_message_set_thread_parent(message, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, param_id, pspec);
		break;
	}
}

static void
xeme_message_init(XemeMessage *message) {
	message->subjects = g_hash_table_new_full(g_str_hash, g_str_equal, g_free,
	                                          g_free);
	message->subjects_order = g_ptr_array_new();

	message->bodies = g_hash_table_new_full(g_str_hash, g_str_equal, g_free,
	                                        g_free);
	message->bodies_order = g_ptr_array_new();
}

static void
xeme_message_class_init(XemeMessageClass *klass) {
	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
	XemeStanzaClass *stanza_class = XEME_STANZA_CLASS(klass);

	obj_class->finalize = xeme_message_finalize;
	obj_class->get_property = xeme_message_get_property;
	obj_class->set_property = xeme_message_set_property;

	stanza_class->marshal = xeme_message_marshal;

	/**
	 * XemeMessage:thread:
	 *
	 * The thread identifier for this message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_THREAD] = g_param_spec_string(
		"thread", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	/**
	 * XemeMessage:thread-parent:
	 *
	 * The thread parent identifier for this message.
	 *
	 * Since: 0.1
	 */
	properties[PROP_THREAD_PARENT] = g_param_spec_string(
		"thread-parent", NULL, NULL,
		NULL,
		G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);

	g_object_class_install_properties(obj_class, N_PROPERTIES, properties);
}

/******************************************************************************
 * Public API
 *****************************************************************************/
XemeStanza *
xeme_message_new(void) {
	return g_object_new(XEME_TYPE_MESSAGE, NULL);
}

const char *
xeme_message_get_thread(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	return message->thread;
}

void
xeme_message_set_thread(XemeMessage *message, const char *thread) {
	g_return_if_fail(XEME_IS_MESSAGE(message));

	if(g_set_str(&message->thread, thread)) {
		g_object_notify_by_pspec(G_OBJECT(message), properties[PROP_THREAD]);
	}
}

const char *
xeme_message_get_thread_parent(XemeMessage *message) {
	g_return_val_if_fail(XEME_IS_MESSAGE(message), NULL);

	return message->thread_parent;
}

void
xeme_message_set_thread_parent(XemeMessage *message,
                               const char *thread_parent)
{
	g_return_if_fail(XEME_IS_MESSAGE(message));

	if(g_set_str(&message->thread_parent, thread_parent)) {
		g_object_notify_by_pspec(G_OBJECT(message),
		                         properties[PROP_THREAD_PARENT]);
	}
}

void
xeme_message_add_body(XemeMessage *message, const char *language,
                      const char *content)
{
	g_return_if_fail(XEME_IS_MESSAGE(message));
	g_return_if_fail(content != NULL);

	xeme_message_add_element_with_language(message, message->bodies,
	                                       message->bodies_order, language,
	                                       content);
}

void
xeme_message_add_subject(XemeMessage *message, const char *language,
                         const char *content)
{
	g_return_if_fail(XEME_IS_MESSAGE(message));
	g_return_if_fail(content != NULL);

	xeme_message_add_element_with_language(message, message->subjects,
	                                       message->subjects_order, language,
	                                       content);
}
